@@ -1,12 +1,13 @@
 /**
- * Copyright (c) 2002-2009 IBM Corporation and others.
+ * Copyright (c) 2002-2014 IBM Corporation, CEA, and others.
  * All rights reserved.   This program and the accompanying materials
- * are made available under the terms of the Eclipse Public License v1.0
+ * are made available under the terms of the Eclipse Public License v2.0
  * which accompanies this distribution, and is available at
- * http://www.eclipse.org/legal/epl-v10.html
+ * http://www.eclipse.org/legal/epl-v20.html
  * 
  * Contributors: 
  *   IBM - Initial API and implementation
+ *   Christian W. Damus (CEA) - 433108
  */
 package org.eclipse.emf.ecore.impl;
 
@@ -61,6 +62,7 @@ import org.eclipse.emf.ecore.util.InternalEList;
  * <!-- end-user-doc -->
  * <p>
  * The following features are implemented:
+ * </p>
  * <ul>
  *   <li>{@link org.eclipse.emf.ecore.impl.EClassImpl#isAbstract <em>Abstract</em>}</li>
  *   <li>{@link org.eclipse.emf.ecore.impl.EClassImpl#isInterface <em>Interface</em>}</li>
@@ -79,7 +81,6 @@ import org.eclipse.emf.ecore.util.InternalEList;
  *   <li>{@link org.eclipse.emf.ecore.impl.EClassImpl#getEGenericSuperTypes <em>EGeneric Super Types</em>}</li>
  *   <li>{@link org.eclipse.emf.ecore.impl.EClassImpl#getEAllGenericSuperTypes <em>EAll Generic Super Types</em>}</li>
  * </ul>
- * </p>
  *
  * @generated
  */
@@ -106,6 +107,11 @@ public class EClassImpl extends EClassifierImpl implements EClass, ESuperAdapter
   protected Map<EOperation, EOperation> eOperationToOverrideMap;
   protected ESuperAdapter eSuperAdapter;
 
+  /**
+   * @since 2.9
+   */
+  protected EGenericType[] eAllStructuralFeatureTypes;
+
   /**
    * <!-- begin-user-doc -->
    * <!-- end-user-doc -->
@@ -143,6 +149,10 @@ public class EClassImpl extends EClassifierImpl implements EClass, ESuperAdapter
         freeze(eOperations.get(i));
       }
     }
+
+    // Bug 433108: Lock in the shared extended metadata for this class
+    ExtendedMetaData.INSTANCE.getName(this);
+
     super.freeze();
   }
   
@@ -218,7 +228,7 @@ public class EClassImpl extends EClassifierImpl implements EClass, ESuperAdapter
           
           protected EClass unwrap(EGenericType eGenericType)
           {
-            EClassifier result = eGenericType.getERawType();
+            EClassifier result = ((EGenericTypeImpl)eGenericType).basicGetERawType();
             if (result instanceof EClass)
             {
               return (EClass)result;
@@ -796,8 +806,8 @@ public class EClassImpl extends EClassifierImpl implements EClass, ESuperAdapter
       {
         private static final long serialVersionUID = 1L;
 
-        protected EStructuralFeature [] containments;
-        protected EStructuralFeature [] crossReferences;
+        protected EStructuralFeature [] containments = NO_EALL_STRUCTURE_FEATURES_DATA;
+        protected EStructuralFeature [] crossReferences = NO_EALL_STRUCTURE_FEATURES_DATA;
 
         public EAllStructuralFeaturesList(BasicEList<EStructuralFeature> eAllStructuralFeatures)
         {
@@ -859,7 +869,7 @@ public class EClassImpl extends EClassifierImpl implements EClass, ESuperAdapter
 
         public EStructuralFeature [] containments()
         {
-          if (containments == null)
+          if (containments == NO_EALL_STRUCTURE_FEATURES_DATA)
           {
             init();
           }
@@ -868,7 +878,7 @@ public class EClassImpl extends EClassifierImpl implements EClass, ESuperAdapter
 
         public EStructuralFeature [] crossReferences()
         {
-          if (crossReferences == null)
+          if (crossReferences == NO_EALL_STRUCTURE_FEATURES_DATA)
           {
             init();
           }
@@ -917,7 +927,7 @@ public class EClassImpl extends EClassifierImpl implements EClass, ESuperAdapter
 
     return eAllStructuralFeatures;
   }
-  
+
   private static final EStructuralFeature[] NO_EALL_STRUCTURE_FEATURES_DATA = {};
 
   /**
@@ -1246,6 +1256,8 @@ public class EClassImpl extends EClassifierImpl implements EClass, ESuperAdapter
         return getOperationID((EOperation)arguments.get(0));
       case EcorePackage.ECLASS___GET_OVERRIDE__EOPERATION:
         return getOverride((EOperation)arguments.get(0));
+      case EcorePackage.ECLASS___GET_FEATURE_TYPE__ESTRUCTURALFEATURE:
+        return getFeatureType((EStructuralFeature)arguments.get(0));
     }
     return eDynamicInvoke(operationID, arguments);
   }
@@ -1298,7 +1310,12 @@ public class EClassImpl extends EClassifierImpl implements EClass, ESuperAdapter
       Map<String, EStructuralFeature> result = new HashMap<String, EStructuralFeature>(3 * eAllStructuralFeatures.size() / 2 + 1);
       for (EStructuralFeature eStructuralFeature : eAllStructuralFeatures)
       {
-        result.put(eStructuralFeature.getName(), eStructuralFeature);
+        String key = eStructuralFeature.getName();
+        EStructuralFeature duplicate = result.put(key, eStructuralFeature);
+        if (duplicate != null)
+        {
+          result.put(key, duplicate);
+        }
       }
       eNameToFeatureMap = result;
     }
@@ -1494,6 +1511,37 @@ public class EClassImpl extends EClassifierImpl implements EClass, ESuperAdapter
     return -1;
   }
 
+  /**
+   * <!-- begin-user-doc -->
+   * <!-- end-user-doc -->
+   * @generated NOT
+   */
+  public EGenericType getFeatureType(EStructuralFeature feature)
+  {
+    int featureID = getFeatureID(feature);
+    if (featureID != -1)
+    {
+      if (eAllStructuralFeatureTypes == null)
+      {
+        EGenericType[] result = new EGenericType[eAllStructuralFeaturesData.length];
+        for (int i = 0; i < eAllStructuralFeaturesData.length; ++i)
+        {
+          EGenericType eGenericType = eAllStructuralFeaturesData[i].getEGenericType();
+          if (eGenericType != null)
+          {
+            result[i] = EcoreUtil.getReifiedType(this, eGenericType);
+          }
+        }
+        eAllStructuralFeatureTypes = result;
+      }
+      return eAllStructuralFeatureTypes[featureID];
+    }
+    else
+    {
+      return feature.getEGenericType();
+    }
+  }
+
   /**
    * The default value of the '{@link #isAbstract() <em>Abstract</em>}' attribute.
    * <!-- begin-user-doc -->
@@ -1650,7 +1698,7 @@ public class EClassImpl extends EClassifierImpl implements EClass, ESuperAdapter
   {
     if (eIsProxy()) return super.toString();
 
-    StringBuffer result = new StringBuffer(super.toString());
+    StringBuilder result = new StringBuilder(super.toString());
     result.append(" (abstract: ");
     result.append((eFlags & ABSTRACT_EFLAG) != 0);
     result.append(", interface: ");
